@@ -31,6 +31,7 @@ import org.apache.myfaces.tobago.event.SortActionEvent;
 import org.apache.myfaces.tobago.internal.component.AbstractUIColumn;
 import org.apache.myfaces.tobago.internal.component.AbstractUIColumnBase;
 import org.apache.myfaces.tobago.internal.component.AbstractUIColumnNode;
+import org.apache.myfaces.tobago.internal.component.AbstractUIColumnPanel;
 import org.apache.myfaces.tobago.internal.component.AbstractUIColumnSelector;
 import org.apache.myfaces.tobago.internal.component.AbstractUIData;
 import org.apache.myfaces.tobago.internal.component.AbstractUILink;
@@ -46,11 +47,14 @@ import org.apache.myfaces.tobago.internal.renderkit.CommandMap;
 import org.apache.myfaces.tobago.internal.util.HtmlRendererUtils;
 import org.apache.myfaces.tobago.internal.util.JsonUtils;
 import org.apache.myfaces.tobago.internal.util.RenderUtils;
+import org.apache.myfaces.tobago.internal.util.StringUtils;
 import org.apache.myfaces.tobago.layout.ShowPosition;
 import org.apache.myfaces.tobago.layout.TextAlign;
 import org.apache.myfaces.tobago.model.ExpandedState;
 import org.apache.myfaces.tobago.model.Selectable;
 import org.apache.myfaces.tobago.model.SheetState;
+import org.apache.myfaces.tobago.model.SortedColumn;
+import org.apache.myfaces.tobago.model.SortedColumnList;
 import org.apache.myfaces.tobago.model.TreePath;
 import org.apache.myfaces.tobago.renderkit.RendererBase;
 import org.apache.myfaces.tobago.renderkit.css.BootstrapClass;
@@ -63,7 +67,6 @@ import org.apache.myfaces.tobago.renderkit.html.HtmlAttributes;
 import org.apache.myfaces.tobago.renderkit.html.HtmlButtonTypes;
 import org.apache.myfaces.tobago.renderkit.html.HtmlElements;
 import org.apache.myfaces.tobago.renderkit.html.HtmlInputTypes;
-import org.apache.myfaces.tobago.util.AjaxUtils;
 import org.apache.myfaces.tobago.util.ComponentUtils;
 import org.apache.myfaces.tobago.util.ResourceUtils;
 import org.apache.myfaces.tobago.webapp.TobagoResponseWriter;
@@ -89,6 +92,7 @@ import java.util.Collections;
 import java.util.List;
 import java.util.Locale;
 import java.util.Map;
+import java.util.Objects;
 
 public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
 
@@ -132,6 +136,9 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
       ComponentUtils.setAttribute(component, Attributes.selectedListString, selectedRows);
     }
 
+    final String sourceId = facesContext.getExternalContext().getRequestParameterMap().get("javax.faces.source");
+    component.setLazyUpdate(sourceId != null && sourceId.equals(clientId + SUFFIX_LAZY));
+
     final String value
         = facesContext.getExternalContext().getRequestParameterMap().get(clientId + SUFFIX_SCROLL_POSITION);
     if (value != null) {
@@ -164,7 +171,6 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
     }
   }
 
-
   private void decodeSheetAction(final FacesContext facesContext, final AbstractUISheet component) {
     final String sourceId = facesContext.getExternalContext().getRequestParameterMap().get("javax.faces.source");
 
@@ -275,7 +281,7 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
     writer.writeAttribute(CustomAttributes.ROWS, component.getRows());
     writer.writeAttribute(CustomAttributes.ROW_COUNT, Integer.toString(component.getRowCount()), false);
     writer.writeAttribute(CustomAttributes.LAZY, component.isLazy());
-    writer.writeAttribute(CustomAttributes.LAZY_UPDATE, component.isLazy() && AjaxUtils.isAjaxRequest(facesContext));
+    writer.writeAttribute(CustomAttributes.LAZY_UPDATE, component.getLazyUpdate());
 
     final boolean autoLayout = component.isAutoLayout();
     if (!autoLayout) {
@@ -402,9 +408,11 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
             writer.writeNameAttribute(pagerCommandId);
             writer.writeAttribute(HtmlAttributes.TYPE, HtmlInputTypes.TEXT);
             writer.writeAttribute(HtmlAttributes.VALUE, first);
+            final int maxLength = Integer.toString(component.getRowCount()).length();
             if (!unknown) {
-              writer.writeAttribute(HtmlAttributes.MAXLENGTH, Integer.toString(component.getRowCount()).length());
+              writer.writeAttribute(HtmlAttributes.MAXLENGTH, maxLength);
             }
+            writer.writeAttribute(HtmlAttributes.PATTERN, StringUtils.positiveNumberRegexp(maxLength), true);
             writer.endElement(HtmlElements.INPUT);
             writer.writeText(formatted.substring(pos + inputMarker.length()));
           } else {
@@ -442,7 +450,7 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
 
       // show page range
       final ShowPosition showPositionPageRange = component.getShowPageRange();
-      if (showPositionRowRange != ShowPosition.none) {
+      if (showPositionPageRange != ShowPosition.none) {
         final AbstractUILink command
             = ensurePagingCommand(facesContext, component, Facets.pagerPage.name(), SheetAction.toPage.name(), false);
         final String pagerCommandId = command.getClientId(facesContext);
@@ -655,17 +663,21 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
         }
       }
       // the row client id depends from the existence of an UIRow component! TBD: is this good?
-      writer.writeIdAttribute(row != null ? row.getClientId(facesContext): sheet.getRowClientId());
+      writer.writeIdAttribute(row != null ? row.getClientId(facesContext) : sheet.getRowClientId());
       writer.writeClassAttribute(
           selected ? TobagoClass.SELECTED : null,
           selected ? BootstrapClass.TABLE_INFO : null,
           row != null ? row.getCustomClass() : null,
           sheet.isRowVisible() ? null : BootstrapClass.D_NONE);
 
+      int colSpan = 0;
+      AbstractUIColumnPanel panel = null;
+
       for (final AbstractUIColumnBase column : columns) {
         if (column.isRendered()) {
           if (column instanceof AbstractUIColumn || column instanceof AbstractUIColumnSelector
               || column instanceof AbstractUIColumnNode) {
+            colSpan++;
             writer.startElement(HtmlElements.TD);
             Markup markup = column.getMarkup();
             if (markup == null) {
@@ -703,10 +715,13 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
             }*/
 
             writer.endElement(HtmlElements.TD);
+          } else if (column instanceof AbstractUIColumnPanel) {
+            panel = (AbstractUIColumnPanel) column;
           }
         }
       }
 
+      colSpan++;
       writer.startElement(HtmlElements.TD);
       writer.startElement(HtmlElements.DIV);
       writer.endElement(HtmlElements.DIV);
@@ -714,6 +729,16 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
       writer.endElement(HtmlElements.TD);
 
       writer.endElement(HtmlElements.TR);
+
+      if (panel != null) {
+        writer.startElement(HtmlElements.TR);
+        writer.writeClassAttribute(panel.getCustomClass());
+        writer.startElement(HtmlElements.TD);
+        writer.writeAttribute(HtmlAttributes.COLSPAN, colSpan);
+        panel.encodeAll(facesContext);
+        writer.endElement(HtmlElements.TD);
+        writer.endElement(HtmlElements.TR);
+      }
     }
 
     sheet.setRowIndex(-1);
@@ -835,6 +860,7 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
             boolean sortable = false;
             boolean ascending = false;
             boolean descending = false;
+            CssItem sortPosition = null;
             String tip = ComponentUtils.getStringAttribute(column, Attributes.tip);
             // sorter icons should only displayed when there is only 1 column and not input
             CommandMap behaviorCommands = null;
@@ -864,16 +890,57 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
                 tip += ResourceUtils.getString(facesContext, "sheet.sorting");
 
                 final SheetState sheetState = sheet.getSheetState(facesContext);
-                if (column.getId().equals(sheetState.getSortedColumnId())) {
-                  final String sortTitle;
-                  if (sheetState.isAscending()) {
-                    sortTitle = ResourceUtils.getString(facesContext, "sheet.ascending");
-                    ascending = true;
-                  } else {
-                    sortTitle = ResourceUtils.getString(facesContext, "sheet.descending");
-                    descending = true;
+                final SortedColumnList sortedColumnList = sheetState.getSortedColumnList();
+
+                final int index = sortedColumnList.indexOf(column.getId());
+                if (index >= 0) {
+                  if (sortedColumnList.isShowNumbers()) { // ignore number circles otherwise
+                    switch (index) {
+                      case 0:
+                        sortPosition = TobagoClass.CIRCLE__1;
+                        break;
+                      case 1:
+                        sortPosition = TobagoClass.CIRCLE__2;
+                        break;
+                      case 2:
+                        sortPosition = TobagoClass.CIRCLE__3;
+                        break;
+                      case 3:
+                        sortPosition = TobagoClass.CIRCLE__4;
+                        break;
+                      case 4:
+                        sortPosition = TobagoClass.CIRCLE__5;
+                        break;
+                      case 5:
+                        sortPosition = TobagoClass.CIRCLE__6;
+                        break;
+                      case 6:
+                        sortPosition = TobagoClass.CIRCLE__7;
+                        break;
+                      case 7:
+                        sortPosition = TobagoClass.CIRCLE__8;
+                        break;
+                      case 8:
+                        sortPosition = TobagoClass.CIRCLE__9;
+                        break;
+                      default:
+                        // should not happen
+                        LOG.warn("Index {} not supported!", index);
+                    }
+                  }
+
+                  final SortedColumn sortedColumn = sortedColumnList.get(index);
+                  if (Objects.equals(column.getId(), sortedColumn.getId())) {
+                    final String sortTitle;
+                    if (sortedColumn.isAscending()) {
+                      sortTitle = ResourceUtils.getString(facesContext, "sheet.ascending");
+                      ascending = true;
+                    } else {
+                      sortTitle = ResourceUtils.getString(facesContext, "sheet.descending");
+                      descending = true;
+                    }
+                    tip += " - " + sortTitle;
                   }
-                  tip += " - " + sortTitle;
                 }
               }
             }
@@ -881,7 +948,8 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
             writer.writeClassAttribute(
                 sortable ? TobagoClass.SORTABLE : null,
                 ascending ? TobagoClass.ASCENDING : null,
-                descending ? TobagoClass.DESCENDING : null);
+                descending ? TobagoClass.DESCENDING : null,
+                sortPosition);
             writer.writeAttribute(HtmlAttributes.TITLE, tip, true);
 
             encodeBehavior(writer, behaviorCommands);
@@ -899,6 +967,13 @@ public class SheetRenderer<T extends AbstractUISheet> extends RendererBase<T> {
               cellComponent.encodeAll(facesContext);
             }
 
+            final UIComponent bar = ComponentUtils.getFacet(column, Facets.bar);
+            if (bar != null) {
+              insideBegin(facesContext, Facets.bar);
+              bar.encodeAll(facesContext);
+              insideEnd(facesContext, Facets.bar);
+            }
+
             writer.endElement(HtmlElements.SPAN);
             if (!autoLayout) {
               if (column.isResizable()) {
